""" Plotting tools for calculations in brightness frame """
import matplotlib.pyplot as plt
import matplotlib.patches
import matplotlib.cm

import numpy as np

import lumos.constants
import lumos.plot
import lumos.conversions
import lumos.geometry

def plot_AB_Mag(observers : lumos.geometry.GroundObservers, 
                intensities : np.ndarray, 
                title : str = None, 
                save_file : str = None, 
                levels : tuple = (0, 8)):
    """
    Creates AB Magnitude plot in brightness frame
        Parameters:
            observers (lumos.geometry.GroundObservers) : Mesh of observers on ground
            intensities (np.ndarray) : Calculated intensities (W / m^2)
            title (str) : Plot title
            save_file (str) : Path to output file
            levels (tuple[int, int]) : Maximum and minimum AB Magnitude to show on plot
    """
    ab_mags = lumos.conversions.intensity_to_ab_mag(intensities)

    with plt.style.context(['dark_background', 'lumos.plot.brightness_frame']):
        fig, ax = plt.subplots()

        max_dist = observers.max_angle * lumos.constants.EARTH_RADIUS / 1000

        ax.set_xlim(-max_dist, max_dist)
        ax.set_ylim(-max_dist, max_dist)
        ax.set_aspect("equal")
    
        ax.set_xlabel("Distance on Plane (km)")
        ax.set_ylabel("Distance off Plane (km)")

        if title != None:
            ax.set_title(title, pad = 15)

        circ = matplotlib.patches.Circle((0, 0), max_dist , transform=ax.transData, facecolor = (1, 1, 1))
        ax.add_patch(circ)

        cmap = matplotlib.cm.get_cmap('plasma_r')

        contour_levels = range(levels[0], levels[-1] + 1)
        ab_mags = np.clip(ab_mags, levels[0] + 0.01, levels[-1] - 0.01)

        cs = ax.contourf(observers.dists_on_plane / 1000,
                        observers.dists_off_plane / 1000,
                        ab_mags,
                        cmap = cmap,
                        levels = contour_levels)
        
        for collection in cs.collections:
            collection.set_clip_path(circ)

        ax.invert_yaxis()
        cbar = fig.colorbar(cs, label = "AB Magnitude")
        cbar.ax.invert_yaxis()

        angles_above_horizon = (15, 30)
        theta = np.linspace(0, 2 * np.pi, 50)

        for angle in angles_above_horizon:
            angle = np.deg2rad(angle)
            R = (lumos.constants.EARTH_RADIUS / 1000
            * (np.pi/2 - angle - np.arcsin(np.cos(observers.max_angle) * np.cos(angle)))
            )

            x = R * np.cos(theta)
            y = R * np.sin(theta)

            annotation_loc = (0.707 * R, 0.707 * R)
            y = np.where( (x - annotation_loc[0])**2 + (y - annotation_loc[1]) ** 2 < 100**2, np.inf, y)

            ax.plot(x, y, '--k')
            ax.annotate(f'{np.rad2deg(angle):.0f}°', annotation_loc, fontsize = 10, color = 'black', 
                    horizontalalignment = 'center', verticalalignment = 'center')
    
        if save_file == None:
            plt.show()
        else:
            fig.savefig(save_file)
            plt.close()


""" Simple calculations that are used throughout Lumos """

import numpy as np

def Rx(theta : float, x : float, y : float, z : float) -> tuple[float, float, float]:
    """
    Rotates points (x, y, z) about x-axis by angle theta

    Parameters:
        theta (float) : Angle of rotation (radians)
        x, y, z (float) : Coordinates of starting point
    
    Returns:
        xp, yp, zp (float) : Coordinates of rotated point
    """
    xp = x
    yp = np.cos(theta) * y - np.sin(theta) * z
    zp = np.sin(theta) * y + np.cos(theta) * z
    return xp, yp, zp

def Ry(theta : float, x : float, y : float, z : float) -> tuple[float, float, float]:
    """
    Rotates points (x, y, z) about y-axis by angle theta

    Parameters:
        theta (float) : Angle of rotation (radians)
        x, y, z (float) : Coordinates of starting point
    
    Returns:
        xp, yp, zp (float) : Coordinates of rotated point
    """
    xp = np.cos(theta) * x + np.sin(theta) * z
    yp = y
    zp = -np.sin(theta) * x + np.cos(theta) * z
    return xp, yp, zp

def Rz(theta : float, x : float, y : float, z : float) -> tuple[float, float, float]:
    """
    Rotates points (x, y, z) about z-axis by angle theta

    Parameters:
        theta (float) : Angle of rotation (radians)
        x, y, z (float) : Coordinates of starting point
    
    Returns:
        xp, yp, zp (float) : Coordinates of rotated point
    """
    xp = np.cos(theta) * x - np.sin(theta) * y
    yp = np.sin(theta) * x + np.cos(theta) * y
    zp = z
    return xp, yp, zp

def det_2(a11, a12,
          a21, a22):
    """
    det(A)

    A = |a11 a12|
        |a21 a22|
    """
    return a11 * a22 - a21 * a12

def det_3(a11, a12, a13,
          a21, a22, a23,
          a31, a32, a33):
    """
    det(A)

    A = |a11 a12 a13|
        |a21 a22 a23|
        |a31 a32 a33|
    """
    return a11 * det_2(a22, a23, a32, a33) - a12 * det_2(a21, a23, a31, a33) + a13 * det_2(a21, a22, a31, a32)

def inv_3(a11, a12, a13,
          a21, a22, a23,
          a31, a32, a33):
    """
    A^-1

    A = |a11 a12 a13|
        |a21 a22 a23|
        |a31 a32 a33|
    """
    c = 1 / det_3(a11, a12, a13,
                  a21, a22, a23,
                  a31, a32, a33)

    I11, I12, I13 = c * det_2(a22, a23, a32, a33), c * det_2(a13, a12, a33, a32), c * det_2(a12, a13, a22, a23)
    I21, I22, I23 = c * det_2(a23, a21, a33, a31), c * det_2(a11, a13, a31, a33), c * det_2(a13, a11, a23, a21)
    I31, I32, I33 = c * det_2(a21, a22, a31, a32), c * det_2(a12, a11, a32, a31), c * det_2(a11, a12, a21, a22)

    return I11, I12, I13, \
           I21, I22, I23, \
           I31, I32, I33


"""
Constants used throughout Lumos
"""

SPEED_OF_LIGHT = 3.0 * 1e8 # meters / second
WAVELENGTH = 532 * 1e-9     # meters
EARTH_RADIUS = 6378 * 1000  # meters
SUN_INTENSITY = 1360 # Watts / meter ^ 2


import os
import cv2

def create_video(image_folder, video_name, frame_rate = 15):
    
    images = [img for img in sorted(os.listdir(image_folder)) if img.endswith(".png")]
    frame = cv2.imread(os.path.join(image_folder, images[0]))
    
    height, width, layers = frame.shape
    
    video = cv2.VideoWriter(video_name, 0, frame_rate, (width, height))
    
    for image in images:
        video.write(cv2.imread(os.path.join(image_folder, image)))
    
    cv2.destroyAllWindows()
    video.release()


"""
Intensity calculations in the brightness reference frame.
"""
import numpy as np
import lumos.constants
import lumos.geometry

def get_earthshine_panels(sat_z : float, angle_past_terminator : float, density : int) \
      -> tuple[np.ndarray, ...]:
    """
    Creates mesh of panels on Earth's surface.

    Parameters:
        sat_z (float): The height of the satellite above geodetic nadir (meters)
        angle_past_terminator (float): The angle of the satellite past terminator (radians)
        density (int): The density of the panels. Grid will have size density x 
    
    Returns:
        x (np.ndarray): x positions of panels (meters)
        y (np.ndarray): y positions of panels (meters)
        z (np.ndarray): z positions of panels (meters)
        nx (np.ndarray): x component of panel normals
        ny (np.ndarray): y component of panel normals
        nz (np.ndarray): z component of panel normals
        areas (np.ndarray): areas of panels (meters ^ 2)
    """
    R = lumos.constants.EARTH_RADIUS

    max_angle = np.arccos(R / sat_z)

    angles_off_plane = np.linspace(-max_angle, max_angle, density)
    angles_on_plane = np.linspace(angle_past_terminator, max_angle, density)

    d_phi = abs( angles_off_plane[1] - angles_off_plane[0] )
    d_theta = abs( angles_on_plane[1] - angles_on_plane[0] )

    angles_on_plane, angles_off_plane = np.meshgrid(angles_on_plane, angles_off_plane)
    angles_on_plane, angles_off_plane = angles_on_plane.flatten(), angles_off_plane.flatten()

    # Set up panel positions
    nz = 1 / np.sqrt( 1 + np.tan(angles_on_plane)**2 + np.tan(angles_off_plane)**2 )
    nx = np.tan(angles_off_plane) * nz
    ny = np.tan(angles_on_plane) * nz

    # Clip the panels which aren't visible to the satellite
    visible_to_sat = np.arccos(nz) < max_angle
    angles_off_plane, angles_on_plane = angles_off_plane[visible_to_sat], angles_on_plane[visible_to_sat]
    nx, ny, nz = nx[visible_to_sat], ny[visible_to_sat], nz[visible_to_sat]

    # Calculate Jacobian determinant to get panel areas
    x, y, z = nx * R, ny * R, nz * R

    phi = angles_off_plane
    theta = angles_on_plane
        
    dx_dr = nx / nz * z / R
    dx_dphi = z**3 / (R**2 * np.cos(phi)**2 * np.cos(theta)**2)
    dx_dtheta = - (ny / nz * nx / nz* z**3 ) / (R**2 * np.cos(theta)**2 )
    
    dy_dr = np.tan(theta) * z / R
    dy_dphi = - ( ny / nz * nx / nz * z**3 ) / (R**2 * np.cos(phi)**2 )
    dy_dtheta = dx_dphi
    
    dz_dr = z / R
    dz_dphi = - (nx / nz * z**3 ) / (R**2 * np.cos(phi)**2)
    dz_dtheta = - (ny / nz * z**3 ) / (R**2 * np.cos(theta)**2)

    determinant = (
        dx_dr * (dy_dphi * dz_dtheta - dy_dtheta * dz_dphi) -
        dy_dr * (dx_dphi * dz_dtheta - dx_dtheta * dz_dphi) +
        dz_dr * (dx_dphi * dy_dtheta - dx_dtheta * dy_dphi) )
    
    areas = determinant * d_phi * d_theta
    
    return x, y, z, nx, ny, nz, areas

def calculate_intensity(sat_surfaces : list[lumos.geometry.Surface], 
                        sat_altitude : float, 
                        angle_past_terminator : float,
                        observer_position : float,
                        include_sun : bool = True, 
                        include_earthshine : bool = True,
                        earth_panel_density : int = 150, 
                        earth_brdf : callable = None) -> float:
    '''
        Parameters:
            sat_surfaces (list[lumos.geometry.Surface]) : A list of lumos.geometry.general.Surface objects
            sat_altitude (float) : Height of satellite above geodetic nadir (meters)
            angle_past_terminator (float) : Angle of satellite past terminator (radians)
            include_sun (bool) : Whether to include brightness directly reflected from sun
            include_earthshine (bool) : Whether to include brightness caused by earthshine
            earth_panel_density (int) : There will be earth_panel_density^2 panels in the earthshine mesh
            earth_brdf (callable) : A function representing the BRDF of earth
        
        Returns:
            intensity (float) : The intensity of light incident on the observer (W / m^2)
    '''
    
    horizon_angle = np.arccos(lumos.constants.EARTH_RADIUS / (lumos.constants.EARTH_RADIUS + sat_altitude))
    if angle_past_terminator > horizon_angle:
        # Inside earth's shadow
        return 0
    
    observer_x, observer_y, observer_z = observer_position[0], observer_position[1], observer_position[2]

    if np.arccos(observer_z / lumos.constants.EARTH_RADIUS) > horizon_angle + np.deg2rad(1):
        # Not visible to observer
        # Fudging this by 1 degree makes plots much nicer on edges
        return 0

    vector_2_sun = (0, np.cos(angle_past_terminator), - np.sin(angle_past_terminator))

    sat_z = sat_altitude + lumos.constants.EARTH_RADIUS

    # Distances from observers to satellite
    dist_sat_2_obs = np.sqrt( observer_x**2 + observer_y**2 
                            + (observer_z - sat_z)**2 )
    
    # Unit vectors from observers to satellite
    sat_obs_x = observer_x / dist_sat_2_obs
    sat_obs_y = observer_y / dist_sat_2_obs
    sat_obs_z = (observer_z - sat_z) / dist_sat_2_obs

    if include_earthshine:
        panel_x, panel_y, panel_z, panel_nx, panel_ny, panel_nz, panel_areas = \
            get_earthshine_panels(sat_z, angle_past_terminator, earth_panel_density)
        
        # Distances from earthshine panels to satellite
        dist_panels_2_sat = np.sqrt(panel_x**2 + panel_y**2 + (sat_z - panel_z)**2 )

        panel_sat_x = - panel_x / dist_panels_2_sat
        panel_sat_y = - panel_y / dist_panels_2_sat
        panel_sat_z = (sat_z - panel_z) / dist_panels_2_sat

        # Panel Normalization
        panel_normalizations = (panel_ny * vector_2_sun[1] + panel_nz * vector_2_sun[2])

        panel_brdfs = earth_brdf(vector_2_sun,
                                (panel_nx, panel_ny, panel_nz),
                                (panel_sat_x, panel_sat_y, panel_sat_z))

    intensity = 0

    for surface in sat_surfaces:
        
        if callable(surface.normal):
            surface.normal = surface.normal(angle_past_terminator)
        
        sun_contribution = 0
        earth_contribution = 0
        
        surface_normalization = surface.normal[1] * vector_2_sun[1] + surface.normal[2] * vector_2_sun[2]

        if include_sun:
            surface_brdf = surface.brdf(vector_2_sun,
                                                surface.normal,
                                                (sat_obs_x, sat_obs_y, sat_obs_z))

            sun_contribution = sun_contribution + surface_brdf * surface_normalization

        if include_earthshine:
            surface_normalizations = -( surface.normal[0] * panel_sat_x 
                             + surface.normal[1] * panel_sat_y
                             + surface.normal[2] * panel_sat_z)

            surface_brdf = surface.brdf( (-panel_sat_x, -panel_sat_y, -panel_sat_z),
                                                   surface.normal,
                                                   (sat_obs_x, sat_obs_y, sat_obs_z) )
            
            earth_contribution = earth_contribution \
                  + np.sum(panel_brdfs * surface_brdf 
                         * panel_normalizations * surface_normalizations
                         * panel_areas / dist_panels_2_sat**2 )
        
        intensity = intensity + lumos.constants.SUN_INTENSITY * surface.area * (sun_contribution + earth_contribution) / dist_sat_2_obs ** 2

    return intensity


"""
Intensity calculations in an observer's reference frame.
"""

import numpy as np
import lumos.calculators.brightness_frame as brightness_frame
import lumos.constants
import lumos.functions
import lumos.conversions
import lumos.geometry

def get_brightness_coords(
    sat_alt : float | np.ndarray, 
    sat_az : float | np.ndarray, 
    sat_height : float | np.ndarray, 
    sun_alt : float, 
    sun_az : float) \
    -> tuple[float | np.ndarray, ...]:
    '''
    Converts from HCS observer coordinates to brightness coordinates.
        Parameters:
            sat_alt (float or np.ndarray) : Altitude of satellite (degrees)
            sat_az (float or np.ndarray) : Azimuth of satellite (degrees)
            sat_height (float or np.ndarray) : Geodetic height of satellite (meters)
            sun_alt (float) : Altitude of sun (degrees)
            sun_az (float) : Azimuth of sun (degrees)
        Returns:
            obs_x (np.ndarray) : x-component of vector from satellite to observer in brightness frame (meters)
            obs_y (np.ndarray) : y-component of vector from satellite to observer in brightness frame (meters)
            obs_z (np.ndarray) : z-component of vector from satellite to observer in brightness frame (meters)
            angle_past_terminator (np.ndarray) : Angle of satellite past the terminator (radians)
    '''

    sat_x, sat_y, sat_z = lumos.conversions.altaz_to_unit(sat_alt, sat_az)
    sun_x, sun_y, sun_z = lumos.conversions.altaz_to_unit(sun_alt, sun_az)

    phi = np.arccos(sat_z) - np.arcsin(lumos.constants.EARTH_RADIUS / (lumos.constants.EARTH_RADIUS + sat_height) * np.sqrt(sat_x**2 + sat_y**2) )
    theta = np.arctan2(sat_y, sat_x)

    Zx = np.sin(phi) * np.cos(theta)
    Zy = np.sin(phi) * np.sin(theta)
    Zz = np.cos(phi)

    dot = Zx * sun_x + Zy * sun_y + Zz * sun_z
    beta = 1 / np.sqrt(1 - dot**2)
    alpha = - beta * dot

    Yx = alpha * Zx + beta * sun_x
    Yy = alpha * Zy + beta * sun_y
    Yz = alpha * Zz + beta * sun_z

    flip = np.sign(Yx * sun_x + Yy * sun_y + Yz * sun_z)
    Yx, Yy, Yz = flip * Yx, flip * Yy, flip * Yz

    Xx = Yy * Zz - Zy * Yz
    Xy = Zx * Yz - Yx * Zz
    Xz = Yx * Zy - Zx * Yy

    T11, T12, T13, \
    T21, T22, T23, \
    T31, T32, T33  = lumos.functions.inv_3(Xx, Yx, Zx, 
                                           Xy, Yy, Zy, 
                                           Xz, Yz, Zz)
    
    dist_to_sat = np.sqrt(lumos.constants.EARTH_RADIUS**2 
    + (lumos.constants.EARTH_RADIUS + sat_height)**2 
    - 2 * lumos.constants.EARTH_RADIUS * (lumos.constants.EARTH_RADIUS + sat_height) * np.cos(phi))

    obs_x = - dist_to_sat * (T11 * sat_x + T12 * sat_y + T13 * sat_z)
    obs_y = - dist_to_sat * (T21 * sat_x + T22 * sat_y + T23 * sat_z)
    obs_z = (lumos.constants.EARTH_RADIUS + sat_height) - dist_to_sat * (T31 * sat_x + T32 * sat_y + T33 * sat_z)

    angle_past_terminator = -np.arcsin(T31 * sun_x + T32 * sun_y + T33 * sun_z)

    return obs_x, obs_y, obs_z, angle_past_terminator

def calculate_intensity(
    sat_surfaces : list[lumos.geometry.Surface], 
    sat_heights : np.ndarray | float, 
    sat_altitudes : np.ndarray | float, 
    sat_azimuths : np.ndarray | float,
    sun_altitude : float, 
    sun_azimuth : float, 
    **kwargs) -> np.ndarray | float:
    '''
    Calculates the intensity of a satellite as seen in a ground observer's reference frame.
        Parameters:
            sat_surfaces (list[Surface]) : A list of lumos.geometry.general.Surface objects
            sat_heights (np.ndarray or float) : Heights of satellites above geodetic nadir (meters)
            sat_altitudes (np.ndarray or float) : Altitude of satellites (degrees)
            sat_azimuths (np.ndarray or float) : Azimuths of satellites (degrees)
            sun_altitude (float) : Altitude of sun
            sun_azimuth (float) : Azimuth of sun
        Keyword Arguments:
            include_sun (bool) : Whether to include brightness directly reflected from sun
            include_earthshine (bool) : Whether to include brightness caused by earthshine
            earth_panel_density (int) : There will be earth_panel_density^2 panels in the earthshine mesh
            earth_brdf (callable) : A function representing the BRDF of earth
        Returns:
            intensity (np.ndarray) : The intensity of light incident on the observer (W / m^2)
    '''

    if sun_altitude > 0:
        raise ValueError("Observatory is in daylight!")
    
    sat_alts, sat_azs = np.asarray(sat_altitudes), np.asarray(sat_azimuths)
    output_shape = sat_alts.shape

    if not isinstance(sat_heights, np.ndarray):
        sat_heights = sat_heights * np.ones(output_shape)

    sat_heights, sat_alts, sat_azs = sat_heights.flatten(), sat_alts.flatten(), sat_azs.flatten()
    intensity = np.zeros_like(sat_alts)

    obs_x, obs_y, obs_z, angle_past_terminator = get_brightness_coords(sat_alts, sat_azs, sat_heights, 
        sun_altitude, sun_azimuth)

    for (i, (x, y, z, alpha, h)) in enumerate(zip(obs_x, obs_y, obs_z, angle_past_terminator, sat_heights)):
        intensity[i] = brightness_frame.calculate_intensity(sat_surfaces, h, alpha, (x, y, z), **kwargs)
    
    intensity = intensity.reshape(output_shape)
    return intensity


""" Conversions which are used throughout Lumos """

import numpy as np
import lumos.constants
import lumos.functions

def intensity_to_ab_mag(intensity : float | np.ndarray, clip : bool = True) -> float | np.ndarray:
    """
    Converts intensity to AB Magnitude

    Parameters:
        intensity (np.ndarray or float) : Intensity measured in watts / meter^2
    
    Returns:
        ab_mag (np.ndarray or float) : Brightness in AB Magnitude
    """
    log_val = intensity * lumos.constants.WAVELENGTH / (lumos.constants.SPEED_OF_LIGHT * 3631e-26)
    if clip:
        log_val = np.clip(log_val, 10e-6, None)
    ab_mag = -2.5 * np.log10( log_val )
    return ab_mag

def altaz_to_unit(altitude : float | np.ndarray, azimuth : float | np.ndarray) \
    -> tuple[ float | np.ndarray, ...]:
    """
    Converts altitude and azimuth to a unit vector.

    Parameters:
        altitude (float or np.ndarray) : Altitude in HCS frame (degrees)
        azimuth (float or np.ndarray) : Azimuth in HCS frame (degrees)
    
    Returns:
        x, y, z (tuple) : Unit vector components
    """
    phi = np.deg2rad(90 - altitude)
    theta = np.deg2rad(azimuth)
    x = np.sin(phi) * np.cos(theta)
    y = np.sin(phi) * np.sin(theta)
    z = np.cos(phi)
    return x, y, z

def unit_to_spherical(x : float | np.ndarray, y : float | np.ndarray, z : float | np.ndarray) \
    -> tuple[float | np.ndarray]:
    """
    Converts a unit vector to spherical coordinates.

    Parameters:
        x (float or np.ndarray)
        y (float or np.ndarray)
        z (float or np.ndarray)
    
    Returns:
        phi, theta (tuple) : Spherical coordinate representation
    """
    phi = np.arccos(z)
    theta = np.arctan2(y, x)
    phi = np.rad2deg(phi)
    theta = np.rad2deg(theta)
    return phi, theta

def spherical_to_unit(phi : float | np.ndarray, theta : float | np.ndarray) \
    -> tuple[float | np.ndarray]:
    """
    Converts a unit vector to spherical coordinates.

    Parameters:
        phi (float or np.ndarray)
        theta (float or np.ndarray)
    
    Returns:
        x, y, z (tuple) : Unit vector representation
    """
    x = np.sin(phi) * np.cos(theta)
    y = np.sin(phi) * np.sin(theta)
    z = np.cos(phi)
    return x, y, z


"""
Geometry objects which are used throughout Lumos
"""
import numpy as np
import lumos.constants
class Surface:
    """ Object to hold area, normal vector, and BRDF of a surface """
    def __init__(self, area : float, normal : np.ndarray, brdf : callable):
        """
        Parameters:
            area (float) : Area of surface (meters ^ 2)
            normal (np.ndarray) : Normal vector of surface. Measured in brightness frame.
            brdf (callable) : BRDF of surface
        """
        self.area = area
        self.normal = normal
        self.brdf = brdf
    
    def __str__(self):
        return f'| Surface \n' \
               f'|-- Area: {self.area:.2f} m^2 \n' \
               f'|-- Normal Vector: ' \
               f'<{self.normal[0]:.2f}, {self.normal[1]:.2f}, {self.normal[2]:.2f}> \n'

class EarthMesh:
    """
    A mesh of points on Earth's surface
    """
    def __init__(self, angles_off_plane : np.ndarray, angles_on_plane : np.ndarray):
        """
        Parameters:
            angles_off_plane (np.ndarray) : 1D array of the angles-off-plane of the mesh
            angles_on_plane (np.ndarray) : 1D array of the angles-on-plane of the mesh
        """
        self.d_phi = np.abs(angles_off_plane[1] - angles_off_plane[0])
        self.d_theta = np.abs(angles_on_plane[1] - angles_on_plane[0])
        
        self.angles_off_plane, self.angles_on_plane = np.meshgrid(angles_off_plane,
                                                                  angles_on_plane)
        
        self.dists_off_plane = lumos.constants.EARTH_RADIUS * self.angles_off_plane
        self.dists_on_plane = lumos.constants.EARTH_RADIUS * self.angles_on_plane
        
        # Set up panel positions
        self.z = ( lumos.constants.EARTH_RADIUS 
                  / np.sqrt(1 + np.tan(self.angles_on_plane)**2 
                            + np.tan(self.angles_off_plane)**2 ) )
                  
        self.x = np.tan(self.angles_off_plane) * self.z
        self.y = np.tan(self.angles_on_plane) * self.z
        
        # Get normal vectors of panels
        self.nx = self.x / lumos.constants.EARTH_RADIUS
        self.ny = self.y / lumos.constants.EARTH_RADIUS
        self.nz = self.z / lumos.constants.EARTH_RADIUS

        self.shape = self.z.shape
        
class GroundObservers(EarthMesh):
    """
    A mesh of observers visible to the satellite and on the night side of earth
    """

    def __init__(self, sat_height : float, angle_past_terminator : float, density : int):
        """
        Parameters:
            sat_height (float) : Geodetic height of satellite (meters)
            angle_past_terminator (float) : Angle of satellite past the terminator (radians)
            density (int) : Mesh will have size density x density
        """
        self.max_angle = np.arccos(lumos.constants.EARTH_RADIUS 
                                   / (lumos.constants.EARTH_RADIUS + sat_height))
        self.min_angle = angle_past_terminator
        
        angles_off_plane = np.linspace(-self.max_angle, self.max_angle, density)
        angles_on_plane = np.linspace(-self.max_angle, self.min_angle, density)
        
        super().__init__(angles_off_plane, angles_on_plane)
    
    def __iter__(self):
        for i in range(self.shape[0]):
            for j in range(self.shape[1]):
                yield i, j, (self.x[i, j], self.y[i, j], self.z[i, j])


import numpy as np

def LAMBERTIAN(TIS : float = 1.0) -> callable:
    """
    Lambertian scattering model

    Parameters:
        TIS (float) : Total integrated scatter, should be between 0.0 and 1.0
    Returns:
        BRDF (callable) : BRDF function which can be used to create a lumos.geometry.Surface object
    """
    def BRDF(incident_vector, normal_vector, outgoing_vector):
        ix, iy, iz = incident_vector
        nx, ny, nz = normal_vector
        ox, oy, oz = outgoing_vector

        brdf = np.where(ix*nx + iy*ny + iz*nz < 0, 0, TIS / np.pi)
        brdf = np.where(ox*nx + oy*ny + oz*nz < 0, 0, TIS / np.pi)

        return brdf
    
    return BRDF

def ABG(A : float = 0.1, B : float = 0.1, g : float = 1.0) -> callable:
    """
    ABg scattering model

    Parameters:
        A (float) : Amplitude parameter, A/B is the specular peak
        B (float) : Knee parameter, B determines where the BRDF transitions to exponential decay
        g (float) : Slope parameter, g determines the slope of the BRDF in log-log space
    Returns:
        BRDF (callable) : BRDF function which can be used to create a lumos.geometry.Surface object
    """
    def BRDF(incident_vector, normal_vector, outgoing_vector):
        ix, iy, iz = incident_vector
        nx, ny, nz = normal_vector
        ox, oy, oz = outgoing_vector
        
        dot = ix*nx + iy*ny + iz*nz
        rx = 2 * dot * nx - ix
        ry = 2 * dot * ny - iy
        rz = 2 * dot * nz - iz

        dot = ox*nx + oy*ny + oz*nz
        beta_x = ox - dot * nx
        beta_y = oy - dot * ny
        beta_z = oz - dot * nz

        dot = rx*nx + ry*ny + rz*nz
        beta0_x = rx - dot * nx
        beta0_y = ry - dot * ny
        beta0_z = rz - dot * nz

        x = np.sqrt( (beta_x - beta0_x)**2 + (beta_y - beta0_y)**2 + (beta_z - beta0_z)**2 )

        brdf = A / (B + x**g)

        brdf = np.where(ix*nx + iy*ny + iz*nz < 0, 0, brdf)
        brdf = np.where(ox*nx + oy*ny + oz*nz < 0, 0, brdf)
        
        return brdf
    
    return BRDF

def GAUSSIAN(A = 1.0, sigma = 1.0) -> callable:
    """
    Gaussian scattering model

    Parameters:
        A (float) : Specular peak of BRDF
        sigma (float) : Width of specular peak
    Returns:
        BRDF (callable) : BRDF function which can be used to create a lumos.geometry.Surface object
    """
    def BRDF(incident_vector, normal_vector, outgoing_vector):
        ix, iy, iz = incident_vector
        nx, ny, nz = normal_vector
        ox, oy, oz = outgoing_vector
        
        dot = ix*nx + iy*ny + iz*nz
        rx = 2 * dot * nx - ix
        ry = 2 * dot * ny - iy
        rz = 2 * dot * nz - iz

        dot = rx * ox + ry * oy + rz * oz
        brdf = A * np.exp( (dot - 1) / sigma )

        brdf = np.where(ix*nx + iy*ny + iz*nz < 0, 0, brdf)
        brdf = np.where(ox*nx + oy*ny + oz*nz < 0, 0, brdf)
        
        return brdf
    
    return BRDF

def PHONG(Kd : float = 0.5, Ks : float = 0.5, n : float = 1) -> callable:
    """
    Phong BRDF model

    Parameters:
        Kd (float) : Diffuse coefficient. Ensure Kd + Ks = 1
        Ks (float) : Specular coefficient. Ensure Kd + Ks = 1
        n (int) : Specularity. Controls width of specular peak.
    Returns:
        BRDF (callable) : BRDF function which can be used to create a lumos.geometry.Surface object
    """
    def BRDF(incident_vector, normal_vector, outgoing_vector):
        ix, iy, iz = incident_vector
        nx, ny, nz = normal_vector
        ox, oy, oz = outgoing_vector
        
        dot = ix*nx + iy*ny + iz*nz
        rx = 2 * dot * nx - ix
        ry = 2 * dot * ny - iy
        rz = 2 * dot * nz - iz

        dot = rx*ox + ry*oy + rz*oz
        dot = np.clip(dot, 0, 1)

        brdf = Kd / np.pi + Ks * (n + 2) / (2 * np.pi) * dot**n

        brdf = np.where(ix*nx + iy*ny + iz*nz < 0, 0, brdf)
        brdf = np.where(ox*nx + oy*ny + oz*nz < 0, 0, brdf)
        
        return brdf
    
    return BRDF

def BINOMIAL(B : np.ndarray, C : np.ndarray, d : float, l1 : int) -> callable:
    """
    Binomial scattering model proposed by Greynolds.

    Parameters:
        B (np.ndarray) : b_i model coefficients
        C (np.ndarray) : c_i model coefficients
        d (np.ndarray) : specularity constant
        l1 (int) : Minimum Gaussian index
    Returns:
        BRDF (callable) : BRDF function which can be used to create a lumos.geometry.Surface object
    """

    def BRDF(incident_vector, normal_vector, outgoing_vector):
        ix, iy, iz = incident_vector
        nx, ny, nz = normal_vector
        ox, oy, oz = outgoing_vector

        dot = ix*nx + iy*ny + iz*nz
        rx = 2 * dot * nx - ix
        ry = 2 * dot * ny - iy
        rz = 2 * dot * nz - iz

        dot = ox*nx + oy*ny + oz*nz
        rho_x = ox - dot * nx
        rho_y = oy - dot * ny
        rho_z = oz - dot * nz

        dot = rx*nx + ry*ny + rz*nz
        rho0_x = rx - dot * nx
        rho0_y = ry - dot * ny
        rho0_z = rz - dot * nz

        D = np.sqrt( (rho_x - rho0_x)**2 + (rho_y - rho0_y)**2 + (rho_z - rho0_z)**2 )
        V = rho_x * rho0_x + rho_y * rho0_y + rho_z * rho0_z

        log_brdf = np.zeros_like(D)
        
        n, m = B.shape
        _, j = C.shape

        for k in range(n):
            term_1 = np.zeros_like(D)

            for i in range(m):
                term_1 = term_1 + B[k, i] * D ** i
            
            term_2 = np.zeros_like(D)
            for i in range(j):
                term_2 = term_2 + C[k, i] * np.log10(1 + d**(i + l1) * D**2)
            
            log_brdf = log_brdf + (term_1 + 0.5 * term_2) * V**k
        
        brdf = 10**log_brdf

        brdf = np.where(ix*nx + iy*ny + iz*nz < 0, 0, brdf)
        brdf = np.where(ox*nx + oy*ny + oz*nz < 0, 0, brdf)
        
        return brdf
    
    return BRDF

class BinomialHelper:
    def __init__(self, n, m, l1, l2):
        self.n = n
        self.m = m
        self.l1 = l1
        self.l2 = l2
        self.N_params = n * m + (l2 - l1) * n + 1

    def pack_params(self, *params):
        params = np.array(params)
        B = np.reshape( params[:self.n * self.m], (self.n, self.m) )
        C = np.reshape( params[self.n * self.m:-1], (self.n, self.l2 - self.l1))
        d = np.abs(params[-1])
        return B, C, d


import numpy as np
import matplotlib.pyplot as plt
import lumos.brdf.library as brdf_library
import lumos.conversions

def plot1D(model, incident_angles = (10, 30, 50, 70)):
    with plt.style.context(["dark_background"]):
        for angle in incident_angles:
            x = np.linspace(-90, 90, 400)
            ix, iy, iz = lumos.conversions.spherical_to_unit(np.deg2rad(angle), np.pi)
            ox, oy, oz = lumos.conversions.spherical_to_unit(np.deg2rad(x), 0)
            y = model((ix, iy, iz), (0, 0, 1), (ox, oy, oz))
            plt.semilogy(x, y, '--k')
        plt.legend()
        plt.show()

def plot2D(model, incident_angles = (10, 30, 50, 70)):
    fig, axs = plt.subplots(ncols = len(incident_angles), subplot_kw={'projection': 'polar'})
    with plt.style.context(["dark_background"]):
        for ax, aoi in zip(axs, incident_angles):
            phi_out = np.linspace(0, 90, 180)
            theta_out = np.linspace(0, 360, 360)
            phi_out, theta_out = np.meshgrid(phi_out, theta_out)

            ix, iy, iz = lumos.conversions.spherical_to_unit(np.deg2rad(aoi), np.pi)
            ox, oy, oz = lumos.conversions.spherical_to_unit(np.deg2rad(phi_out), np.deg2rad(theta_out))
            y = model((ix, iy, iz), (0, 0, 1), (ox, oy, oz))
            ax.contourf(np.deg2rad(theta_out), phi_out, np.log10(y))
            ax.set_title(r"$\phi_{in}$ = " + f"{aoi:0.1f}°")
            ax.set_xticks([])
            ax.set_yticks([])
            ax.set_theta_zero_location('N')
        plt.show()


"""
Tools for fitting experimental BRDF data to models
"""

import numpy as np
import scipy.optimize
import lumos.conversions

def fit_model(
        data_file : str, 
        model_func : callable, 
        p0 : tuple[float, ...], 
        bounds : tuple[float, ...],
        log_space : bool = True,
        clip : float = 1e-4
        ) -> tuple[float, ...]:
    
    """
    Fits a model to experimental data.

    Parameters:
        data_file (str) : File containing experimental data.
        model_func (callable) : Given parameters *params, returns BRDF callable
        p0 (tuple) : Initial guess for fitting parameters, passed to scipy.optimize.curve_fit
        bounds (tuple) : Bounds for fitting parameters, passed to scipy.optimize.curve_fit
        clip (float) : Clips BRDF data below this value
    
    Returns:
        popt (tuple) : Optimal parameters returned from scipy.optimize.curve_fit
    """

    data = np.loadtxt(data_file, skiprows = 1)

    phi_in = np.deg2rad(data[:, 0])
    theta_in = np.deg2rad(data[:, 1])
    phi_out = np.deg2rad(data[:, 2])
    theta_out = np.deg2rad(data[:, 3])
    brdf = data[:, 4]

    mask = brdf > clip

    phi_in = phi_in[mask]
    theta_in = theta_in[mask]
    phi_out = phi_out[mask]
    theta_out = theta_out[mask]
    brdf = brdf[mask]

    indexes = np.arange(brdf.size)

    ix, iy, iz = lumos.conversions.spherical_to_unit(phi_in, theta_in)
    ox, oy, oz = lumos.conversions.spherical_to_unit(phi_out, theta_out)

    def fit_function(idx, *params):
        model_brdf = model_func(*params)

        idx = idx.astype(int)

        f = model_brdf(
            (ix[idx], iy[idx], iz[idx]),
            (0, 0, 1),
            (ox[idx], oy[idx], oz[idx])
            )
        
        return np.log10(f) if log_space else f

    popt, _ = scipy.optimize.curve_fit(fit_function, 
                                       indexes, 
                                       np.log10(brdf) if log_space else brdf, 
                                       bounds = bounds,
                                       p0 = p0)

    return popt


""" Plotting AB Magnitude in ground frame """

import matplotlib.pyplot as plt
import matplotlib.cm
import matplotlib.gridspec
import numpy as np
import lumos.plot
import lumos.conversions as conversions
import scipy.ndimage

def plot_AB_Mag_contour(altitudes : np.ndarray, 
                        azimuths : np.ndarray, 
                        intensities : np.ndarray, 
                        sun_alt : float, 
                        sun_az : float,
                        levels : tuple[int, int] = (0, 8),
                        filter : bool = True, 
                        title : str = None, 
                        save_file : str = None):
    """
    Creates AB Magnitude plot in ground frame
        Parameters:
            altitudes (np.ndarray) : Altitudes (degrees)
            azimuths (np.ndarray) : Azimuths (degrees)
            intensities (np.ndarray) : Calculated intensities (W / m^2)
            sun_alt (float) : Altitude of sun (degrees)
            sun_az (float) : Azimuth of sun (degrees)
            levels (tuple[int, int]) : Maximum and minimum AB Magnitude to show on plot
            title (str) : Plot title
            save_file (str) : Path to output file
    """
    
    if filter:
        intensities = scipy.ndimage.gaussian_filter(intensities, 2, mode = ('wrap', 'reflect'))
    
    ab_mags = conversions.intensity_to_ab_mag(intensities)

    with plt.style.context(['dark_background', 'lumos.plot.ground_frame']):
        fig = plt.figure()

        gs = matplotlib.gridspec.GridSpec(1, 3, width_ratios = [1, 5, 1])
        cax1 = plt.subplot(gs[0])
        ax = plt.subplot(gs[1], projection = 'polar')
        cax2 = plt.subplot(gs[2])

        plt.subplots_adjust(top=0.75)

        # -------------- Central Heat Map ----------------------
        if title != None:
            fig.suptitle(title, x = 0.5, y = 0.95)
        
        ab_mags = np.clip(ab_mags, levels[0] + 0.01, levels[-1] - 0.01)

        cmap = matplotlib.colormaps['plasma_r']
        cs = ax.contourf(np.radians(azimuths), 90 - altitudes,
                         ab_mags, cmap = cmap, 
                         levels = range(levels[0], levels[-1] + 1))
        
        ax.plot([np.deg2rad(sun_az)], [95], marker = (3, 0, 180 - sun_az), markersize = 6, color = "white", clip_on = False)
        ax.plot([np.deg2rad(sun_az)], [101], marker = '$\u2600$', markersize = 10, color = "orange", clip_on = False)

        ax.set_rmax(90)
        ax.set_yticklabels([])
        ax.set_theta_zero_location('N')
        ax.set_theta_direction(-1)
        ax.set_rticks([10, 20, 30, 40, 50, 60, 70, 80, 90])
        ax.set_xticks(np.deg2rad([0, 90, 180, 270]))
        ax.set_xticklabels(['N', 'E', 'S', 'W'])
        ax.set_rlabel_position(-22.5)
        ax.grid(True)

        plt.colorbar(cs, label = "AB Magnitude", cax = cax1, location = 'left')
        cax1.invert_yaxis()
        cax1.set_aspect(20)

        evening_cmap = matplotlib.colors.LinearSegmentedColormap.from_list('Evening', ['#15171c', '#20242d', '#25406a', '#4872bc', '#88a5d1', '#b5c9e6'])
        norm = matplotlib.colors.Normalize(vmin=-18, vmax=0)
        plt.colorbar(matplotlib.cm.ScalarMappable(norm = norm, cmap = evening_cmap), cax = cax2)
        cax2.set_aspect('equal')
        cax2.set_xlim(0, 1)
        cax2.yaxis.set_tick_params(which = 'minor', length = 0, rotation = 0, pad = 35)
        cax2.yaxis.set_tick_params(which = 'major', length = 5, pad = 5)
        cax2.xaxis.set_label_position("top")
        cax2.set_yticks(ticks = [-18, -12, -6, 0], labels = ['Night', '', '', 'Day'], fontsize = 12)
        cax2.set_yticks(ticks = [-15.5, -9.5, -3.5], labels = ['Astronomical\nTwilight', 'Nautical\nTwilight', 'Civil\nTwilight'], minor = True, fontsize = 8, ma = 'center', ha = 'center')
        plot_alt = np.clip(sun_alt, -18, 0)
        cax2.plot([-0.4], [plot_alt], marker = (3, 0, 270), color = 'white', markersize = 6, clip_on = False)
        cax2.plot([-1.0], [plot_alt], marker = '$\u2600$', color = 'orange', markersize = 10, clip_on = False)
    
        if save_file == None:
            plt.show()
        else:
            fig.savefig(save_file)
            plt.close()

def plot_AB_Mag_scatter(altitudes : np.ndarray, 
                        azimuths : np.ndarray, 
                        intensities : np.ndarray, 
                        sun_alt : float, 
                        sun_az : float,
                        levels : tuple[int, int] = (0, 8),
                        title : str = None, 
                        save_file : str = None):
    """
    Creates AB Magnitude plot in ground frame
        Parameters:
            altitudes (np.ndarray) : Altitudes (degrees)
            azimuths (np.ndarray) : Azimuths (degrees)
            intensities (np.ndarray) : Calculated intensities (W / m^2)
            sun_alt (float) : Altitude of sun (degrees)
            sun_az (float) : Azimuth of sun (degrees)
            levels (tuple[int, int]) : Maximum and minimum AB Magnitude to show on plot
            title (str) : Plot title
            save_file (str) : Path to output file
    """
    
    ab_mags = conversions.intensity_to_ab_mag(intensities)

    with plt.style.context(['dark_background', 'lumos.plot.ground_frame']):
        fig = plt.figure()

        gs = matplotlib.gridspec.GridSpec(1, 3, width_ratios = [1, 5, 1])
        cax1 = plt.subplot(gs[0])
        ax = plt.subplot(gs[1], projection = 'polar')
        cax2 = plt.subplot(gs[2])

        plt.subplots_adjust(top=0.75)

        # -------------- Central Scatter Plot ----------------------
        if title != None:
            fig.suptitle(title, x = 0.5, y = 0.95)
        
        ab_mags = np.clip(ab_mags, levels[0] + 0.01, levels[-1] - 0.01)

        cmap = matplotlib.colormaps['plasma_r']
        cs = ax.scatter(np.radians(azimuths), 90 - altitudes,
            c = ab_mags, cmap = cmap, 
            norm = matplotlib.colors.Normalize(vmin = levels[0], vmax = levels[-1]))
        
        ax.plot([np.deg2rad(sun_az)], [95], marker = (3, 0, 180 - sun_az), markersize = 6, color = "white", clip_on = False)
        ax.plot([np.deg2rad(sun_az)], [101], marker = '$\u2600$', markersize = 10, color = "orange", clip_on = False)

        ax.set_rmax(90)
        ax.set_yticklabels([])
        ax.set_theta_zero_location('N')
        ax.set_theta_direction(-1)
        ax.set_rticks([10, 20, 30, 40, 50, 60, 70, 80, 90])
        ax.set_xticks(np.deg2rad([0, 90, 180, 270]))
        ax.set_xticklabels(['N', 'E', 'S', 'W'])
        ax.set_rlabel_position(-22.5)
        ax.grid(True)

        plt.colorbar(cs, label = "AB Magnitude", cax = cax1, location = 'left')
        cax1.invert_yaxis()
        cax1.set_xlim((0, 1))
        cax1.set_aspect(18 / (levels[1] - levels[0]))

        evening_cmap = matplotlib.colors.LinearSegmentedColormap.from_list('Evening', ['#15171c', '#20242d', '#25406a', '#4872bc', '#88a5d1', '#b5c9e6'])
        norm = matplotlib.colors.Normalize(vmin=-18, vmax=0)
        plt.colorbar(matplotlib.cm.ScalarMappable(norm = norm, cmap = evening_cmap), cax = cax2)
        cax2.set_aspect('equal')
        cax2.set_xlim(0, 1)
        cax2.yaxis.set_tick_params(which = 'minor', length = 0, rotation = 0, pad = 35)
        cax2.yaxis.set_tick_params(which = 'major', length = 5, pad = 5)
        cax2.xaxis.set_label_position("top")
        cax2.set_yticks(ticks = [-18, -12, -6, 0], labels = ['Night', '', '', 'Day'], fontsize = 12)
        cax2.set_yticks(ticks = [-15.5, -9.5, -3.5], labels = ['Astronomical\nTwilight', 'Nautical\nTwilight', 'Civil\nTwilight'], minor = True, fontsize = 8, ma = 'center', ha = 'center')
        plot_alt = np.clip(sun_alt, -18, 0)
        cax2.plot([-0.4], [plot_alt], marker = (3, 0, 270), color = 'white', markersize = 6, clip_on = False)
        cax2.plot([-1.0], [plot_alt], marker = '$\u2600$', color = 'orange', markersize = 10, clip_on = False)
    
        if save_file == None:
            plt.show()
        else:
            fig.savefig(save_file)
            plt.close()